package ch.psi.daq.queryrest.config;

import java.util.Arrays;
import java.util.EnumMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Function;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;

import org.msgpack.jackson.dataformat.MessagePackFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.util.StringUtils;
import org.springframework.validation.Validator;

import com.fasterxml.jackson.annotation.JsonInclude.Include;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.smile.SmileFactory;

import ch.psi.daq.common.statistic.Statistics;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.config.DomainConfigCORS;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.query.operation.aggregation.extrema.AbstractExtremaMeta;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.query.analyzer.BackendQueryAnalyzerImpl;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.queryrest.controller.validator.QueryValidator;
import ch.psi.daq.queryrest.model.PropertyFilterMixin;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.query.QueryManagerImpl;
import ch.psi.daq.queryrest.response.PolymorphicResponseMixIn;
import ch.psi.daq.queryrest.response.csv.CSVResponseStreamWriter;
import ch.psi.daq.queryrest.response.json.JSONResponseStreamWriter;
import ch.psi.daq.queryrest.response.json.JSONTableResponseStreamWriter;
import ch.psi.daq.queryrest.response.msgpack.MsgPackResponseStreamWriter;
import ch.psi.daq.queryrest.response.msgpack.MsgPackTableResponseStreamWriter;
import ch.psi.daq.queryrest.response.smile.SmileResponseStreamWriter;
import ch.psi.daq.queryrest.response.smile.SmileTableResponseStreamWriter;

@Configuration
@Import(value = DomainConfigCORS.class)
@PropertySource(value = {"classpath:queryrest.properties"})
@PropertySource(value = {
      "classpath:queryrest-${daq.config.environment}.properties",
      "file:${user.home}/.config/daq/queryrest.properties"}, ignoreResourceNotFound = true)
public class QueryRestConfig { // extends WebMvcConfigurerAdapter {

   private static final String QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS = "queryrest.default.response.aggregations";

   private static final String QUERYREST_DEFAULT_RESPONSE_FIELDS = "queryrest.default.response.fields";

   // a nested configuration
   // this guarantees that the ordering of the properties file is as expected
   // see:
   // https://jira.spring.io/browse/SPR-10409?focusedCommentId=101393&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-101393
   @Configuration
   @Import({QueryConfig.class})
   static class InnerConfiguration {
   }

   // somehow needed to make sure @Import elements will get initialized and afterPropertiesSet will
   // get called (ensuring BackendAcces gets initialized according hierarchy)
   @Resource
   private QueryConfig queryConfig;

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestConfig.class);

   public static final String BEAN_NAME_QUERY_MANAGER = "queryManager";
   public static final String BEAN_NAME_QUERY_ANALIZER_FACTORY = "queryAnalizerFactory";
   public static final String BEAN_NAME_QUERY_VALIDATOR = "queryValidator";
   public static final String BEAN_NAME_JSON_FACTORY = "jsonFactory";
   public static final String BEAN_NAME_MSG_PACK_FACTORY = "msgPackFactory";
   public static final String BEAN_NAME_SMILE_FACTORY = "smileFactory";
   public static final String BEAN_NAME_DEFAULT_RESPONSE_FIELDS = "defaultResponseFields";
   public static final String BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS = "defaultResponseAggregations";

   @Resource
   private ApplicationContext context;
   private Environment env;
   private ObjectMapper objectMapper;

   @SuppressWarnings("unchecked")
   @PostConstruct
   public void afterPropertiesSet() {
      env = context.getEnvironment();
      objectMapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);

      initObjectMapper(objectMapper);

      // init BackendAccesses
      final List<Backend> backends = context.getBean(DomainConfig.BEAN_NAME_BACKENDS, List.class);
      for (final Backend backend : backends) {
         // backends instance is different to objectMapper above
         initObjectMapper(
               backend.getApplicationContext().getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class));
      }
   }

   private void initObjectMapper(ObjectMapper objectMapper) {
      // only include non-null values
      objectMapper.setSerializationInclusion(Include.NON_NULL);
      // Mixin which is used dynamically to filter out which properties get serialised and which
      // won't. This way, the user can specify which columns are to be received.
      objectMapper.addMixIn(DataEvent.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(Statistics.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(AbstractExtremaMeta.class, PropertyFilterMixin.class);
      objectMapper.addMixIn(EnumMap.class, PropertyFilterMixin.class);

      objectMapper.addMixIn(Response.class, PolymorphicResponseMixIn.class);
   }

//   @Override
//   public void configureMessageConverters(List<HttpMessageConverter<?>> converters) {
//      final MappingJackson2HttpMessageConverter converter = new MappingJackson2HttpMessageConverter();
//      /**
//       * This is necessary so that the message conversion uses the configured object mapper.
//       * Otherwise, a separate object mapper is instantiated for Springs message conversion.
//       */
//      converter.setObjectMapper(objectMapper);
//      converters.add(converter);
//      super.configureMessageConverters(converters);
//   }
   
//   does this work for json comming from web-requests
//   -> use WebMvcConfigurationSupport?
//         https://stackoverflow.com/questions/26639475/how-to-set-context-param-in-spring-boot
//   @Bean
//   @Lazy
//   public MappingJackson2HttpMessageConverter mappingJackson2HttpMessageConverter() {
//      final MappingJackson2HttpMessageConverter jsonConverter = new MappingJackson2HttpMessageConverter();
//      jsonConverter.setObjectMapper(objectMapper);
//      return jsonConverter;
//   }

   @Bean(name = BEAN_NAME_JSON_FACTORY)
   @Lazy
   public JsonFactory jsonFactory() {
      return new JsonFactory();
   }

   @Bean(name = BEAN_NAME_MSG_PACK_FACTORY)
   @Lazy
   public MessagePackFactory messagePackFactory() {
      return new MessagePackFactory();
   }

   @Bean(name = BEAN_NAME_SMILE_FACTORY)
   @Lazy
   public SmileFactory smileFactory() {
      return new SmileFactory();
   }

   @Bean(name = BEAN_NAME_QUERY_ANALIZER_FACTORY)
   @Lazy
   public Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory() {
      return (query) -> new BackendQueryAnalyzerImpl(query);
   }

   @Bean
   @Lazy
   public JSONResponseStreamWriter jsonResponseStreamWriter() {
      return new JSONResponseStreamWriter();
   }

   @Bean
   @Lazy
   public JSONTableResponseStreamWriter jsonTableResponseStreamWriter() {
      return new JSONTableResponseStreamWriter();
   }

   @Bean
   @Lazy
   public MsgPackResponseStreamWriter msgPackResponseStreamWriter() {
      return new MsgPackResponseStreamWriter();
   }

   @Bean
   @Lazy
   public MsgPackTableResponseStreamWriter msgPackTableResponseStreamWriter() {
      return new MsgPackTableResponseStreamWriter();
   }

   @Bean
   @Lazy
   public SmileResponseStreamWriter smileResponseStreamWriter() {
      return new SmileResponseStreamWriter();
   }

   @Bean
   @Lazy
   public SmileTableResponseStreamWriter smileTableResponseStreamWriter() {
      return new SmileTableResponseStreamWriter();
   }

   @Bean
   @Lazy
   public CSVResponseStreamWriter csvResponseStreamWriter() {
      return new CSVResponseStreamWriter();
   }

   @Bean(name = BEAN_NAME_QUERY_MANAGER)
   @Lazy
   public QueryManager queryManager() {
      return new QueryManagerImpl();
   }

   @Bean(name = BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   @Lazy
   public Set<QueryField> defaultResponseFields() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_DEFAULT_RESPONSE_FIELDS));
      LOGGER.debug("Load '{}={}'", BEAN_NAME_DEFAULT_RESPONSE_FIELDS, Arrays.toString(responseFields));

      // preserve order
      LinkedHashSet<QueryField> defaultResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            defaultResponseFields.add(QueryField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in '{}' is invalid.", field, QUERYREST_DEFAULT_RESPONSE_FIELDS, e);
         }
      }

      return defaultResponseFields;
   }

   @Bean(name = BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   @Lazy
   public Set<Aggregation> defaultResponseAggregations() {
      String[] responseAggregations =
            StringUtils.commaDelimitedListToStringArray(env.getProperty(QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS));
      LOGGER.debug("Load '{}={}'", BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS, Arrays.toString(responseAggregations));

      // preserve order
      LinkedHashSet<Aggregation> defaultResponseAggregations = new LinkedHashSet<>(responseAggregations.length);
      for (String aggregation : responseAggregations) {
         try {
            defaultResponseAggregations.add(Aggregation.valueOf(aggregation));
         } catch (Exception e) {
            LOGGER.error("Aggregation '{}' in '{}' is invalid.", aggregation, QUERYREST_DEFAULT_RESPONSE_AGGREGATIONS,
                  e);
         }
      }

      return defaultResponseAggregations;
   }

   @Bean(name = BEAN_NAME_QUERY_VALIDATOR)
   @Lazy
   public Validator queryValidator() {
      return new QueryValidator();
   }
}
