package ch.psi.daq.queryrest.query;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.PreDestroy;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.json.channels.info.ChannelInfos;
import ch.psi.daq.domain.query.ChannelNameRequest;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.backend.BackendQueryImpl;
import ch.psi.daq.domain.query.backend.analyzer.BackendQueryAnalyzer;
import ch.psi.daq.domain.query.channels.ChannelNameCache;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.processor.QueryProcessor;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.query.model.ChannelInfosStreamImpl;

public class QueryManagerImpl implements QueryManager, ApplicationContextAware {
   private ChannelNameCache channelNameCache;
   private Function<BackendQuery, BackendQueryAnalyzer> queryAnalizerFactory;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      channelNameCache = context.getBean(QueryConfig.BEAN_NAME_CHANNEL_NAME_CACHE, ChannelNameCache.class);
      queryAnalizerFactory = context.getBean(QueryRestConfig.BEAN_NAME_QUERY_ANALIZER_FACTORY, Function.class);
   }

   @PreDestroy
   public void destroy() {}

   @Override
   public List<ChannelsResponse> getChannels(ChannelsRequest request) {
      // in case not specified use defaults (e.g. GET)
      if (request == null) {
         request = new ChannelsRequest();
      }

      return channelNameCache.getChannels(request);
   }

   public Collection<ChannelInfos> getChannelInfos(final ChannelNameRequest request) {
      // set backends if not defined yet
      channelNameCache.configureBackends(request.getChannels());

      final Stream<ChannelInfos> stream = request.getRequestsByBackend().entrySet().stream()
            .filter(entry -> entry.getKey().getBackendAccess().hasDataReader()
                  && entry.getKey().getBackendAccess().hasChannelInfoReader())
            .flatMap(entry -> {
               return entry.getValue().getChannelInfos(entry.getKey())
                     .entrySet().stream()
                     .map(innerEntry -> {
                        return new ChannelInfosStreamImpl(
                              new ChannelName(innerEntry.getKey(), entry.getKey()),
                              innerEntry.getValue());
                     });
            });

      // materialize
      return stream.collect(Collectors.toList());
   }

   @Override
   public List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> getEvents(
         final DAQQueries queries) {
      // set backends if not defined yet
      channelNameCache.configureBackends(queries);

      final List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results =
            new ArrayList<>(queries.getQueries().size());

      for (final DAQQueryElement queryElement : queries) {
         Stream<Triple<BackendQuery, ChannelName, ?>> resultStreams =
               BackendQueryImpl
                     .getBackendQueries(queryElement)
                     .stream()
                     .filter(
                           query -> query.getBackend().getBackendAccess().hasDataReader()
                                 && query.getBackend().getBackendAccess().hasQueryProcessor())
                     .flatMap(
                           query -> {
                              final QueryProcessor processor = query.getBackend().getBackendAccess().getQueryProcessor();
                              final BackendQueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);

                              /* all the magic happens here */
                              final Stream<Entry<ChannelName, Stream<? extends DataEvent>>> channelToDataEvents =
                                    processor.process(queryAnalizer);
                              /* do post-process */
                              final Stream<Entry<ChannelName, ?>> channelToData =
                                    queryAnalizer.postProcess(channelToDataEvents);

                              return channelToData.map(entry -> {
                                 return Triple.of(query, entry.getKey(), entry.getValue());
                              });
                           });

         // Now we have a stream that loads elements sequential BackendQuery by BackendQuery.
         // By materializing the outer Stream the elements of all BackendQuery are loaded async
         // (speeds things up but requires also more memory - i.e. it relies on Backends not loading
         // all elements into memory at once)
         resultStreams = resultStreams.collect(Collectors.toList()).stream();

         results.add(Pair.of(queryElement, resultStreams));
      }

      return results;
   }

}
