package ch.psi.daq.queryrest.response.json;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.ToLongFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.servlet.ServletResponse;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;

import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;

import ch.psi.daq.common.stream.match.ListCreator;
import ch.psi.daq.common.stream.match.ListFiller;
import ch.psi.daq.common.stream.match.Padder;
import ch.psi.daq.common.stream.match.StreamMatcher;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.backend.BackendQuery;
import ch.psi.daq.domain.query.bin.BinningStrategy;
import ch.psi.daq.domain.query.bin.strategy.BinningStrategyPerBinPulse;
import ch.psi.daq.domain.query.bin.strategy.BinningStrategyPerBinTime;
import ch.psi.daq.domain.query.mapping.IncompleteStrategy;
import ch.psi.daq.domain.query.mapping.Mapping;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.request.range.RequestRange;
import ch.psi.daq.query.bin.aggregate.BinnedValueCombiner;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

/**
 * Takes a Java 8 stream and writes it to the output stream provided by the {@link ServletResponse}
 * of the current request.
 */
public class JSONTableResponseStreamWriter implements ResponseStreamWriter, ApplicationContextAware {

   private static final Logger LOGGER = LoggerFactory.getLogger(JSONTableResponseStreamWriter.class);

   public static final Mapping DEFAULT_MAPPING = new Mapping(IncompleteStrategy.PROVIDE_AS_IS);
   private static final long MILLIS_PER_PULSE = TimeUtils.MILLIS_PER_PULSE;
   private static final Function<DataEvent, ChannelName> KEY_PROVIDER = (event) -> new ChannelName(event.getChannel(),
         event.getBackend());
   // try to match sync data (bsread) with non sync data (epics) based on the time usin 10 millis
   // buckets.
   private static final ToLongFunction<DataEvent> MATCHER_PROVIDER = (event) -> event.getGlobalMillis()
         / MILLIS_PER_PULSE;


   private ObjectMapper mapper;
   private JsonFactory factory;
   // In case ArchiverAppliance had several events within the 10ms mapping interval, return these
   // aggregations
   private Set<String> defaultResponseAggregationsStr;

   @SuppressWarnings("unchecked")
   @Override
   public void setApplicationContext(ApplicationContext context) throws BeansException {
      final Backend backend = context.getBean(DomainConfig.BEAN_NAME_BACKEND_DEFAULT, Backend.class);
      context = backend.getApplicationContext();

      mapper = context.getBean(DomainConfig.BEAN_NAME_OBJECT_MAPPER, ObjectMapper.class);
      factory = context.getBean(QueryRestConfig.BEAN_NAME_JSON_FACTORY, JsonFactory.class);

      final Set<Aggregation> defaultResponseAggregations =
            context.getBean(QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS, Set.class);;
      defaultResponseAggregationsStr =
            defaultResponseAggregations.stream().map(Aggregation::name)
                  .collect(Collectors.toCollection(LinkedHashSet::new));
   }

   @Override
   public void respond(final List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         final OutputStream out, final Response response) throws Exception {
      respond(factory, mapper, defaultResponseAggregationsStr, results, out, response);
   }


   @SuppressWarnings("unchecked")
   public static void respond(final JsonFactory factory,
         final ObjectMapper mapper, final Set<String> defaultResponseAggregationsStr,
         final List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> results,
         final OutputStream out, final Response response) throws Exception {
      final AtomicReference<Exception> exception = new AtomicReference<>();
      final JsonGenerator generator = factory.createGenerator(out, JsonEncoding.UTF8);

      try {
         if (results.size() > 1) {
            generator.writeStartArray();
         }

         results
               .forEach(entryy -> {
                  final DAQQueryElement daqQuery = entryy.getKey();
                  final Set<String> includedFields = JSONResponseStreamWriter.getFields(daqQuery, false);
                  /* make sure identifiers are available */
                  includedFields.add(QueryField.channel.name());
                  includedFields.add(QueryField.backend.name());
                  // issue ATEST-633
                  if (!containsAggregation(includedFields)) {
                     includedFields.addAll(defaultResponseAggregationsStr);
                  }

                  final ObjectWriter writer = JSONResponseStreamWriter.configureWriter(includedFields, mapper);

                  /* get DataEvent stream of sub-queries for later match */
                  final Map<ChannelName, Stream<DataEvent>> streams =
                        new LinkedHashMap<>(results.size());
                  final AtomicReference<BackendQuery> backendQueryRef = new AtomicReference<>();

                  entryy.getValue()
                        .sequential()
                        .forEach(
                              triple -> {
                                 backendQueryRef.compareAndSet(null, triple.getLeft());

                                 if (triple.getRight() instanceof Stream) {
                                    streams.put(triple.getMiddle(), ((Stream<DataEvent>) triple.getRight()));
                                 } else {
                                    final String message =
                                          String.format("Expect a DataEvent Stream for '%s' but got '%s'.",
                                                triple.getMiddle(), triple.getRight().getClass().getSimpleName());
                                    LOGGER.warn(message);
                                    streams.put(triple.getMiddle(), Stream.empty());
                                 }
                              });

                  final BackendQuery backendQuery = backendQueryRef.get();
                  final RequestRange requestRange = backendQuery.getRequest().getRequestRange();
                  BinningStrategy binningStrategy = backendQuery.getBinningStrategy();

                  final Mapping mapping = daqQuery.getMappingOrDefault(DEFAULT_MAPPING);
                  final Padder<ChannelName, DataEvent> padder = mapping.getIncomplete().getPadder(backendQuery);

                  ToLongFunction<DataEvent> matchProvider = binningStrategy;
                  if (binningStrategy == null) {
                     matchProvider = MATCHER_PROVIDER;
                     if (requestRange.isPulseIdRangeDefined()) {
                        binningStrategy = new BinningStrategyPerBinPulse(1);
                     } else if (requestRange.isTimeRangeDefined()) {
                        binningStrategy = new BinningStrategyPerBinTime(MILLIS_PER_PULSE);
                     } else {
                        final String message = "Either time or pulseId range must be defined by the query!";
                        LOGGER.error(message);
                        throw new IllegalStateException(message);
                     }
                  }
                  binningStrategy.setRequestRange(requestRange);

                  /* online matching of the stream's content */
                  final StreamMatcher<ChannelName, DataEvent, List<DataEvent>> streamMatcher =
                        new StreamMatcher<>(
                              KEY_PROVIDER,
                              matchProvider,
                              new ListCreator<ChannelName, DataEvent>(),
                              new ListFiller<ChannelName, DataEvent>(),
                              new BinnedValueCombiner(binningStrategy),
                              padder,
                              streams.values());
                  final Iterator<List<DataEvent>> streamsMatchIter = streamMatcher.iterator();

                  try {
                     generator.writeStartObject();
                     generator.writeFieldName(JSONResponseStreamWriter.DATA_RESP_FIELD);
                     writer.writeValue(generator, streamsMatchIter);
                     generator.writeEndObject();
                  } catch (Exception e) {
                     LOGGER.error("Exception while writing json for '{}'", daqQuery.getChannels(), e);
                     exception.compareAndSet(null, e);
                  } finally {
                     if (streamMatcher != null) {
                        try {
                           streamMatcher.close();
                        } catch (Throwable t) {
                           LOGGER.error(
                                 "Something went wrong while closing stream matcher for JSON table response writer.",
                                 t);
                        }
                     }
                  }
               });
      } catch (IOException e) {
         LOGGER.error("Could not write JSON.", e);
         exception.compareAndSet(null, e);
      } finally {
         if (results.size() > 1) {
            generator.writeEndArray();
         }

         generator.flush();
         generator.close();
      }

      if (exception.get() != null) {
         throw exception.get();
      }
   }

   private static boolean containsAggregation(final Set<String> includedFields) {
      for (final Aggregation aggregation : Aggregation.values()) {
         if (includedFields.contains(aggregation.name())) {
            return true;
         }
      }
      return false;
   }
}
